'''
File consists of several classes for the different elements of a device.
'''
from __future__ import annotations
import uuid
from typing import Any, Union, Literal, TypedDict, TypeVar, Type, List, Optional, Dict

# TODO
# - Parents direkt setzen
# - Alle gemeinsamen JSON attribute im LegoItem constructor
# - enforce uuid
# - Docstrings
# - Beschreibung von Teilen (-> properties)
# - Gute String Darstellung
# - Minimalbeispiel für KPIs
# - Klassen analog zur deutschen Aufgabenstellung umbenennen (assembly, component) #13 #17
# - Export als JSON
# - Export als GraphViz
# - AggregationsEbene enum (system, assembly, subassembly, component)


class LegoItem:
    def __init__(self, item_number: int, mass: float, delivery_time: int, **kwargs) -> None:
        # , *args, **kwargs not handling additional/optional specs right now
        self.id: uuid.UUID = uuid.uuid4()
        self.properties: dict = kwargs
        self.item_number: int = item_number
        self.mass: float = mass
        self.delivery_time: int = delivery_time
        # TODO: Set parent directly and not via id? This would allow for easier traversal of the tree
        #   Currently there is no way to search for parts and components in tree by id.
        self.parent_id = None  # This will be set when added to a component

    def __str__(self):
        return (
            f"Item(id={self.id}, item_number={self.item_number}, "
            f"mass={self.mass}, delivery_time={self.delivery_time}, "
            f"parent_id={self.parent_id})"
        )

    def __repr__(self):
        return f"Lego Item [{self.id}]"


class LegoComponent:
    def __init__(self, **kwargs) -> None:
        self.id: uuid.UUID = uuid.uuid4()
        self.properties: dict = kwargs
        self.items: List[LegoItem] = []
        self.components: List[LegoComponent] = []
        self.parent_id: None | uuid.UUID = None

    def add_item(self, item: LegoItem) -> None:
        if not isinstance(item, LegoItem):
            raise TypeError(f"'item' should be of type LegoPart, got {type(item).__name__} instead.")
        item.parent_id = self.id
        self.items.append(item)

    def add_component(self, component: LegoComponent) -> None:
        if not isinstance(component, LegoComponent):
            raise TypeError(f"'component' should be of type LegoComponent, got {type(component).__name__} instead.")
        component.parent_id = self.id
        self.components.append(component)

    def children(self) -> Dict[str, List[LegoItem] | List[LegoComponent]]:
        return {'items': self.items, 'components': self.components}

    def get_item_list(self) -> List[LegoItem]:
        item_list = []
        item_list.extend(self.items)
        for component in self.components:
            item_list.extend(component.get_item_list())
        return item_list

    def __repr__(self):
        return f"Lego Component [{self.id}]"

# TODO: Adjust default output when printing an item or component

def print_component_tree(root, level=0, is_last=False):
    # print component
    component_padding = ""
    if level > 0:
        component_padding += "│   " * (level - 1)
        if is_last:
            component_padding += "└── "
        else:
            component_padding += "├── "
    print(f"{component_padding}{root}")
    # recursively print child components
    for i, component in enumerate(root.components):
        is_last_ = i == len(root.components) - 1 and len(root.items) == 0
        print_component_tree(component, level + 1, is_last_)
    # print items
    for i, item in enumerate(root.items):
        item_padding = "│   " * level
        item_padding += "├── " if i < len(root.items) - 1 else "└── "
        print(f"{item_padding}{item}")
